@@ -63,7 +63,7 @@ MOVETEXT_REGEX = re.compile(r"""
     |(\))
     |(\*|1-0|0-1|1/2-1/2)
     |(
-        [NBKRQ]?[a-h]?[1-8]?[\-x]?[a-h][1-8](?:=[nbrqNBRQ])?
+        [NBKRQ]?[a-h]?[1-8]?[\-x]?[a-h][1-8](?:=?[nbrqNBRQ])?
         |--
         |O-O(?:-O)?
         |0-0(?:-0)?
@@ -86,7 +86,7 @@ class GameNode(object):
 
     def board(self):
         """
-        Gets a bitboard with the position of the node.
+        Gets a board with the position of the node.
 
         It's a copy, so modifying the board will not alter the game.
         """
@@ -263,7 +263,7 @@ class GameNode(object):
 
                 # Recursively append the next moves.
                 _board.push(variation.move)
-                variation.export(exporter, comments, variations, _board, _after_variation=False)
+                variation.export(exporter, comments, variations, _board, False)
                 _board.pop()
 
                 # End variation.
@@ -275,7 +275,7 @@ class GameNode(object):
 
             # Recursively append the next moves.
             _board.push(main_variation.move)
-            main_variation.export(exporter, comments, variations, _board, _after_variation=variations and len(self.variations) > 1)
+            main_variation.export(exporter, comments, variations, _board, variations and len(self.variations) > 1)
             _board.pop()
 
     def __str__(self):
@@ -324,7 +324,7 @@ class Game(GameNode):
 
     def board(self):
         """
-        Gets the starting position of the game as a bitboard.
+        Gets the starting position of the game.
 
         Unless the `SetUp` and `FEN` header tags are set this is the default
         starting position.
@@ -363,7 +363,7 @@ class Game(GameNode):
         if comments and self.comment:
             exporter.put_starting_comment(self.comment)
 
-        super(Game, self).export(exporter, comments=comments, variations=variations, _after_variation=True)
+        super(Game, self).export(exporter, comments=comments, variations=variations)
 
         exporter.put_result(self.headers["Result"])
         exporter.end_game()
