@@ -20,6 +20,8 @@ import chess
 import mmap
 import os
 import struct
+import sys
+import threading
 
 
 UINT64_BE = struct.Struct(">Q")
@@ -28,7 +30,7 @@ UINT32_BE = struct.Struct(">I")
 USHORT = struct.Struct("<H")
 
 WDL_MAGIC = [0x71, 0xE8, 0x23, 0x5D]
-DTZ_MAGIC = [0xd7, 0x66, 0x0c, 0xa5]
+DTZ_MAGIC = [0xD7, 0x66, 0x0C, 0xA5]
 
 OFFDIAG = [
     0, -1, -1, -1, -1, -1, -1, -1,
@@ -351,32 +353,14 @@ def calc_key_from_filename(filename, mirror=False):
 def subfactor(k, n):
     f = n
     l = 1
-    i = 1
-    while i < k:
+
+    for i in range(1, k):
         f *= n - i
         l *= i + 1
-        i += 1
 
     return f // l
 
 
-def _bb(board, piece_type, color):
-    if piece_type == chess.PAWN:
-        bb = board.pawns
-    elif piece_type == chess.KNIGHT:
-        bb = board.knights
-    elif piece_type == chess.BISHOP:
-        bb = board.bishops
-    elif piece_type == chess.ROOK:
-        bb = board.rooks
-    elif piece_type == chess.QUEEN:
-        bb = board.queens
-    elif piece_type == chess.KING:
-        bb = board.kings
-
-    return bb & board.occupied_co[color]
-
-
 class PairsData(object):
     def __init__(self):
         self.indextable = None
@@ -417,6 +401,14 @@ class Table(object):
         self.fd = os.open(os.path.join(directory, filename) + suffix, os.O_RDONLY | os.O_BINARY if hasattr(os, "O_BINARY") else os.O_RDONLY)
         self.data = mmap.mmap(self.fd, 0, access=mmap.ACCESS_READ)
 
+        if sys.version_info >= (3, ):
+            self.read_ubyte = self.data.__getitem__
+        else:
+            def read_ubyte(data_ptr):
+                return ord(self.data[data_ptr])
+
+            self.read_ubyte = read_ubyte
+
         self.key = calc_key_from_filename(filename)
         self.mirrored_key = calc_key_from_filename(filename, True)
         self.symmetric = self.key == self.mirrored_key
@@ -495,14 +487,10 @@ class Table(object):
 
         d.base = [0 for _ in range(h)]
         d.base[h - 1] = 0
-        i = h - 2
-        while i >= 0:
+        for i in range(h - 2, -1, -1):
             d.base[i] = (d.base[i + 1] + self.read_ushort(d.offset + i * 2) - self.read_ushort(d.offset + i * 2 + 2)) // 2
-            i -= 1
-        i = 0
-        while i < h:
+        for i in range(h):
             d.base[i] <<= 64 - (min_len + i)
-            i += 1
 
         d.offset -= 2 * d.min_len
 
@@ -597,11 +585,9 @@ class Table(object):
         tmp[s] = 1
 
     def pawn_file(self, pos):
-        i = 1
-        while i < self.pawns[0]:
+        for i in range(1, self.pawns[0]):
             if FLAP[pos[0]] > FLAP[pos[i]]:
                 pos[0], pos[i] = pos[i], pos[0]
-            i += 1
 
         return FILE_TO_FILE[pos[0] & 0x07]
 
@@ -657,28 +643,20 @@ class Table(object):
         while i < n:
             t = norm[i]
 
-            j = i
-            while j < i + t:
-                k = j + 1
-                while k < i + t:
+            for j in range(i, i + t):
+                for k in range(j + 1, i + t):
                     # Swap.
                     if pos[j] > pos[k]:
                         pos[j], pos[k] = pos[k], pos[j]
-                    k += 1
-                j += 1
 
             s = 0
 
-            m = i
-            while m < i + t:
+            for m in range(i, i + t):
                 p = pos[m]
-                l = 0
                 j = 0
-                while l < i:
+                for l in range(i):
                     j += int(p > pos[l])
-                    l += 1
                 s += BINOMIAL[m - i][p - j]
-                m += 1
 
             idx += s * factor[i]
             i += t
@@ -692,71 +670,49 @@ class Table(object):
             for i in range(n):
                 pos[i] ^= 0x07
 
-        i = 1
-        while i < self.pawns[0]:
-            j = i + 1
-            while j < self.pawns[0]:
+        for i in range(1, self.pawns[0]):
+            for j in range(i + 1, self.pawns[0]):
                 if PTWIST[pos[i]] < PTWIST[pos[j]]:
                     pos[i], pos[j] = pos[j], pos[i]
-                j += 1
-            i += 1
 
         t = self.pawns[0] - 1
         idx = PAWNIDX[t][FLAP[pos[0]]]
-        i = t
-        while i > 0:
+        for i in range(t, 0, -1):
             idx += BINOMIAL[t - i][PTWIST[pos[i]]]
-            i -= 1
         idx *= factor[0]
 
         # Remaining pawns.
         i = self.pawns[0]
         t = i + self.pawns[1]
         if t > i:
-            j = i
-            while j < t:
-                k = j + 1
-                while k < t:
+            for j in range(i, t):
+                for k in range(j + 1, t):
                     if pos[j] > pos[k]:
                         pos[j], pos[k] = pos[k], pos[j]
-                    k += 1
-                j += 1
             s = 0
-            m = i
-            while m < t:
+            for m in range(i, t):
                 p = pos[m]
-                k = 0
                 j = 0
-                while k < i:
+                for k in range(i):
                     j += int(p > pos[k])
-                    k += 1
                 s += BINOMIAL[m - i][p - j - 8]
-                m += 1
             idx += s * factor[i]
             i = t
 
         while i < n:
             t = norm[i]
-            j = i
-            while j < i + t:
-                k = j + 1
-                while k < i + t:
+            for j in range(i, i + t):
+                for k in range(j + 1, i + t):
                     if pos[j] > pos[k]:
                         pos[j], pos[k] = pos[k], pos[j]
-                    k += 1
-                j += 1
-            s = 0
 
-            m = i
-            while m < i + t:
+            s = 0
+            for m in range(i, i + t):
                 p = pos[m]
-                k = 0
                 j = 0
-                while k < i:
+                for k in range(i):
                     j += int(p > pos[k])
-                    k += 1
                 s += BINOMIAL[m - i][p - j]
-                m += 1
 
             idx += s * factor[i]
             i += t
@@ -836,9 +792,6 @@ class Table(object):
     def read_ushort(self, data_ptr):
         return USHORT.unpack_from(self.data, data_ptr)[0]
 
-    def read_ubyte(self, data_ptr):
-        return ord(self.data[data_ptr:data_ptr + 1])
-
     def close(self):
         self.data.close()
 
@@ -850,127 +803,140 @@ class Table(object):
     def __enter__(self):
         return self
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, exc_type, exc_value, traceback):
         self.close()
 
     def __getstate__(self):
         state = self.__dict__.copy()
         del state["fd"]
         del state["data"]
+        del state["read_ubyte"]
+        del state["lock"]
         return state
 
     def __setstate__(self, state):
-        self.__dict__.update(state)
         self.__init__(self.directory, self.filename, self.suffix)
+        self.__dict__.update(state)
 
 
 class WdlTable(Table):
 
     def __init__(self, directory, filename, suffix=".rtbw"):
         super(WdlTable, self).__init__(directory, filename, suffix)
-
-        assert WDL_MAGIC[0] == self.read_ubyte(0)
-        assert WDL_MAGIC[1] == self.read_ubyte(1)
-        assert WDL_MAGIC[2] == self.read_ubyte(2)
-        assert WDL_MAGIC[3] == self.read_ubyte(3)
+        self.initialized = False
+        self.lock = threading.Lock()
 
     def init_table_wdl(self):
-        self.tb_size = [0 for _ in range(8)]
-        self.size = [0 for _ in range(8 * 3)]
+        if self.initialized:
+            return
 
-        # Used if there are only pieces.
-        self.precomp = {}
-        self.pieces = {}
+        with self.lock:
+            if self.initialized:
+                return
 
-        self.factor = {}
-        self.factor[chess.WHITE] = [0, 0, 0, 0, 0, 0]
-        self.factor[chess.BLACK] = [0, 0, 0, 0, 0, 0]
+            assert WDL_MAGIC[0] == self.read_ubyte(0)
+            assert WDL_MAGIC[1] == self.read_ubyte(1)
+            assert WDL_MAGIC[2] == self.read_ubyte(2)
+            assert WDL_MAGIC[3] == self.read_ubyte(3)
 
-        self.norm = {}
-        self.norm[chess.WHITE] = [0 for _ in range(self.num)]
-        self.norm[chess.BLACK] = [0 for _ in range(self.num)]
+            self.tb_size = [0 for _ in range(8)]
+            self.size = [0 for _ in range(8 * 3)]
 
-        # Used if there are pawns.
-        self.files = [PawnFileData() for _ in range(4)]
+            # Used if there are only pieces.
+            self.precomp = {}
+            self.pieces = {}
 
-        self._next = None
-        self._flags = None
-        self.flags = None
+            self.factor = {}
+            self.factor[chess.WHITE] = [0, 0, 0, 0, 0, 0]
+            self.factor[chess.BLACK] = [0, 0, 0, 0, 0, 0]
 
-        split = self.read_ubyte(4) & 0x01
-        files = 4 if self.read_ubyte(4) & 0x02 else 1
+            self.norm = {}
+            self.norm[chess.WHITE] = [0 for _ in range(self.num)]
+            self.norm[chess.BLACK] = [0 for _ in range(self.num)]
 
-        data_ptr = 5
+            # Used if there are pawns.
+            self.files = [PawnFileData() for _ in range(4)]
 
-        if not self.has_pawns:
-            self.setup_pieces_piece(data_ptr)
-            data_ptr += self.num + 1
-            data_ptr += data_ptr & 0x01
+            self._next = None
+            self._flags = None
+            self.flags = None
 
-            self.precomp[chess.WHITE] = self.setup_pairs(data_ptr, self.tb_size[0], 0, True)
-            data_ptr = self._next
-            if split:
-                self.precomp[chess.BLACK] = self.setup_pairs(data_ptr, self.tb_size[1], 3, True)
-                data_ptr = self._next
-            else:
-                self.precomp[chess.BLACK] = None
+            split = self.read_ubyte(4) & 0x01
+            files = 4 if self.read_ubyte(4) & 0x02 else 1
 
-            self.precomp[chess.WHITE].indextable = data_ptr
-            data_ptr += self.size[0]
-            if split:
-                self.precomp[chess.BLACK].indextable = data_ptr
-                data_ptr += self.size[3]
+            data_ptr = 5
 
-            self.precomp[chess.WHITE].sizetable = data_ptr
-            data_ptr += self.size[1]
-            if split:
-                self.precomp[chess.BLACK].sizetable = data_ptr
-                data_ptr += self.size[4]
+            if not self.has_pawns:
+                self.setup_pieces_piece(data_ptr)
+                data_ptr += self.num + 1
+                data_ptr += data_ptr & 0x01
 
-            data_ptr = (data_ptr + 0x3f) & ~0x3f
-            self.precomp[chess.WHITE].data = data_ptr
-            data_ptr += self.size[2]
-            if split:
-                data_ptr = (data_ptr + 0x3f) & ~0x3f
-                self.precomp[chess.BLACK].data = data_ptr
-        else:
-            s = 1 + int(self.pawns[1] > 0)
-            for f in range(4):
-                self.setup_pieces_pawn(data_ptr, 2 * f, f)
-                data_ptr += self.num + s
-            data_ptr += data_ptr & 0x01
-
-            for f in range(files):
-                self.files[f].precomp[chess.WHITE] = self.setup_pairs(data_ptr, self.tb_size[2 * f], 6 * f, True)
+                self.precomp[chess.WHITE] = self.setup_pairs(data_ptr, self.tb_size[0], 0, True)
                 data_ptr = self._next
                 if split:
-                    self.files[f].precomp[chess.BLACK] = self.setup_pairs(data_ptr, self.tb_size[2 * f + 1], 6 * f + 3, True)
+                    self.precomp[chess.BLACK] = self.setup_pairs(data_ptr, self.tb_size[1], 3, True)
                     data_ptr = self._next
                 else:
-                    self.files[f].precomp[chess.BLACK] = None
+                    self.precomp[chess.BLACK] = None
 
-            for f in range(files):
-                self.files[f].precomp[chess.WHITE].indextable = data_ptr
-                data_ptr += self.size[6 * f]
+                self.precomp[chess.WHITE].indextable = data_ptr
+                data_ptr += self.size[0]
                 if split:
-                    self.files[f].precomp[chess.BLACK].indextable = data_ptr
-                    data_ptr += self.size[6 * f + 3]
+                    self.precomp[chess.BLACK].indextable = data_ptr
+                    data_ptr += self.size[3]
 
-            for f in range(files):
-                self.files[f].precomp[chess.WHITE].sizetable = data_ptr
-                data_ptr += self.size[6 * f + 1]
+                self.precomp[chess.WHITE].sizetable = data_ptr
+                data_ptr += self.size[1]
                 if split:
-                    self.files[f].precomp[chess.BLACK].sizetable = data_ptr
-                    data_ptr += self.size[6 * f + 4]
+                    self.precomp[chess.BLACK].sizetable = data_ptr
+                    data_ptr += self.size[4]
 
-            for f in range(files):
                 data_ptr = (data_ptr + 0x3f) & ~0x3f
-                self.files[f].precomp[chess.WHITE].data = data_ptr
-                data_ptr += self.size[6 * f + 2]
+                self.precomp[chess.WHITE].data = data_ptr
+                data_ptr += self.size[2]
                 if split:
                     data_ptr = (data_ptr + 0x3f) & ~0x3f
-                    self.files[f].precomp[chess.BLACK].data = data_ptr
-                    data_ptr += self.size[6 * f + 5]
+                    self.precomp[chess.BLACK].data = data_ptr
+            else:
+                s = 1 + int(self.pawns[1] > 0)
+                for f in range(4):
+                    self.setup_pieces_pawn(data_ptr, 2 * f, f)
+                    data_ptr += self.num + s
+                data_ptr += data_ptr & 0x01
+
+                for f in range(files):
+                    self.files[f].precomp[chess.WHITE] = self.setup_pairs(data_ptr, self.tb_size[2 * f], 6 * f, True)
+                    data_ptr = self._next
+                    if split:
+                        self.files[f].precomp[chess.BLACK] = self.setup_pairs(data_ptr, self.tb_size[2 * f + 1], 6 * f + 3, True)
+                        data_ptr = self._next
+                    else:
+                        self.files[f].precomp[chess.BLACK] = None
+
+                for f in range(files):
+                    self.files[f].precomp[chess.WHITE].indextable = data_ptr
+                    data_ptr += self.size[6 * f]
+                    if split:
+                        self.files[f].precomp[chess.BLACK].indextable = data_ptr
+                        data_ptr += self.size[6 * f + 3]
+
+                for f in range(files):
+                    self.files[f].precomp[chess.WHITE].sizetable = data_ptr
+                    data_ptr += self.size[6 * f + 1]
+                    if split:
+                        self.files[f].precomp[chess.BLACK].sizetable = data_ptr
+                        data_ptr += self.size[6 * f + 4]
+
+                for f in range(files):
+                    data_ptr = (data_ptr + 0x3f) & ~0x3f
+                    self.files[f].precomp[chess.WHITE].data = data_ptr
+                    data_ptr += self.size[6 * f + 2]
+                    if split:
+                        data_ptr = (data_ptr + 0x3f) & ~0x3f
+                        self.files[f].precomp[chess.BLACK].data = data_ptr
+                        data_ptr += self.size[6 * f + 5]
+
+            self.initialized = True
 
     def setup_pieces_pawn(self, p_data, p_tb_size, f):
         j = 1 + int(self.pawns[chess.BLACK] > 0)
@@ -1002,6 +968,8 @@ class WdlTable(Table):
         self.tb_size[chess.BLACK] = self.calc_factors_piece(self.factor[chess.BLACK], order, self.norm[chess.BLACK])
 
     def probe_wdl_table(self, board):
+        self.init_table_wdl()
+
         key = calc_key(board)
 
         if self.symmetric:
@@ -1023,7 +991,7 @@ class WdlTable(Table):
             while i < self.num:
                 piece_type = self.pieces[bside][i] & 0x07
                 color = (self.pieces[bside][i] ^ cmirror) >> 3
-                bb = _bb(board, piece_type, color)
+                bb = board.pieces_mask(piece_type, color)
 
                 square = chess.bit_scan(bb)
                 while square != -1 and square is not None:
@@ -1039,7 +1007,7 @@ class WdlTable(Table):
             k = self.files[0].pieces[0][0] ^ cmirror
             color = k >> 3
             piece_type = k & 0x07
-            bb = _bb(board, piece_type, color)
+            bb = board.pieces_mask(piece_type, color)
 
             square = chess.bit_scan(bb)
             while square != -1 and square is not None:
@@ -1052,7 +1020,7 @@ class WdlTable(Table):
             while i < self.num:
                 color = (pc[i] ^ cmirror) >> 3
                 piece_type = pc[i] & 0x07
-                bb = _bb(board, piece_type, color)
+                bb = board.pieces_mask(piece_type, color)
 
                 square = chess.bit_scan(bb)
                 while square != -1 and square is not None:
@@ -1070,86 +1038,99 @@ class DtzTable(Table):
 
     def __init__(self, directory, filename, suffix=".rtbz"):
         super(DtzTable, self).__init__(directory, filename, suffix)
-
-        assert DTZ_MAGIC[0] == self.read_ubyte(0)
-        assert DTZ_MAGIC[1] == self.read_ubyte(1)
-        assert DTZ_MAGIC[2] == self.read_ubyte(2)
-        assert DTZ_MAGIC[3] == self.read_ubyte(3)
+        self.initialized = False
+        self.lock = threading.Lock()
 
     def init_table_dtz(self):
-        self.factor = [0, 0, 0, 0, 0, 0]
-        self.norm = [0 for _ in range(self.num)]
-        self.tb_size = [0, 0, 0, 0]
-        self.size = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
-        self.files = [PawnFileDataDtz() for f in range(4)]
+        if self.initialized:
+            return
 
-        files = 4 if self.read_ubyte(4) & 0x02 else 1
+        with self.lock:
+            if self.initialized:
+                return
 
-        p_data = 5
+            assert DTZ_MAGIC[0] == self.read_ubyte(0)
+            assert DTZ_MAGIC[1] == self.read_ubyte(1)
+            assert DTZ_MAGIC[2] == self.read_ubyte(2)
+            assert DTZ_MAGIC[3] == self.read_ubyte(3)
 
-        if not self.has_pawns:
-            self.map_idx = [0, 0, 0, 0]
+            self.factor = [0, 0, 0, 0, 0, 0]
+            self.norm = [0 for _ in range(self.num)]
+            self.tb_size = [0, 0, 0, 0]
+            self.size = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]
+            self.files = [PawnFileDataDtz() for f in range(4)]
 
-            self.setup_pieces_piece_dtz(p_data, 0)
-            p_data += self.num + 1
-            p_data += p_data & 0x01
+            files = 4 if self.read_ubyte(4) & 0x02 else 1
 
-            self.precomp = self.setup_pairs(p_data, self.tb_size[0], 0, False)
-            self.flags = self._flags
-            p_data = self._next
-            self.p_map = p_data
-            if self.flags & 2:
-                for i in range(4):
-                    self.map_idx[i] = p_data + 1 - self.p_map
-                    p_data += 1 + self.read_ubyte(p_data)
+            p_data = 5
+
+            if not self.has_pawns:
+                self.map_idx = [0, 0, 0, 0]
+
+                self.setup_pieces_piece_dtz(p_data, 0)
+                p_data += self.num + 1
                 p_data += p_data & 0x01
 
-            self.precomp.indextable = p_data
-            p_data += self.size[0]
-
-            self.precomp.sizetable = p_data
-            p_data += self.size[1]
-
-            p_data = (p_data + 0x3f) & ~0x3f
-            self.precomp.data = p_data
-            p_data += self.size[2]
-        else:
-            s = 1 + int(self.pawns[1] > 0)
-            for f in range(4):
-                self.setup_pieces_pawn_dtz(p_data, f, f)
-                p_data += self.num + s
-            p_data += p_data & 0x01
-
-            self.flags = []
-            for f in range(files):
-                self.files[f].precomp = self.setup_pairs(p_data, self.tb_size[f], 3 * f, False)
+                self.precomp = self.setup_pairs(p_data, self.tb_size[0], 0, False)
+                self.flags = self._flags
                 p_data = self._next
-                self.flags.append(self._flags)
-
-            self.map_idx = []
-            self.p_map = p_data
-            for f in range(files):
-                self.map_idx.append([])
-                if self.flags[f] & 2:
+                self.p_map = p_data
+                if self.flags & 2:
                     for i in range(4):
-                        self.map_idx[-1].append(p_data + 1 - self.p_map)
+                        self.map_idx[i] = p_data + 1 - self.p_map
                         p_data += 1 + self.read_ubyte(p_data)
-            p_data += p_data & 0x01
+                    p_data += p_data & 0x01
 
-            for f in range(files):
-                self.files[f].precomp.indextable = p_data
-                p_data += self.size[3 * f]
+                self.precomp.indextable = p_data
+                p_data += self.size[0]
 
-            for f in range(files):
-                self.files[f].precomp.sizetable = p_data
-                p_data += self.size[3 * f + 1]
+                self.precomp.sizetable = p_data
+                p_data += self.size[1]
 
-            for f in range(files):
                 p_data = (p_data + 0x3f) & ~0x3f
-                self.files[f].precomp.data = p_data
-                p_data += self.size[3 * f + 2]
+                self.precomp.data = p_data
+                p_data += self.size[2]
+            else:
+                s = 1 + int(self.pawns[1] > 0)
+                for f in range(4):
+                    self.setup_pieces_pawn_dtz(p_data, f, f)
+                    p_data += self.num + s
+                p_data += p_data & 0x01
+
+                self.flags = []
+                for f in range(files):
+                    self.files[f].precomp = self.setup_pairs(p_data, self.tb_size[f], 3 * f, False)
+                    p_data = self._next
+                    self.flags.append(self._flags)
+
+                self.map_idx = []
+                self.p_map = p_data
+                for f in range(files):
+                    self.map_idx.append([])
+                    if self.flags[f] & 2:
+                        for i in range(4):
+                            self.map_idx[-1].append(p_data + 1 - self.p_map)
+                            p_data += 1 + self.read_ubyte(p_data)
+                p_data += p_data & 0x01
+
+                for f in range(files):
+                    self.files[f].precomp.indextable = p_data
+                    p_data += self.size[3 * f]
+
+                for f in range(files):
+                    self.files[f].precomp.sizetable = p_data
+                    p_data += self.size[3 * f + 1]
+
+                for f in range(files):
+                    p_data = (p_data + 0x3f) & ~0x3f
+                    self.files[f].precomp.data = p_data
+                    p_data += self.size[3 * f + 2]
+
+            self.initialized = True
 
     def probe_dtz_table(self, board, wdl):
+        self.init_table_dtz()
+
         key = calc_key(board)
 
         if not self.symmetric:
@@ -1175,7 +1156,7 @@ class DtzTable(Table):
             while i < self.num:
                 piece_type = pc[i] & 0x07
                 color = (pc[i] ^ cmirror) >> 3
-                bb = _bb(board, piece_type, color)
+                bb = board.pieces_mask(piece_type, color)
 
                 square = chess.bit_scan(bb)
                 while square != -1 and square is not None:
@@ -1195,7 +1176,7 @@ class DtzTable(Table):
             k = self.files[0].pieces[0] ^ cmirror
             piece_type = k & 0x07
             color = k >> 3
-            bb = _bb(board, piece_type, color)
+            bb = board.pieces_mask(piece_type, color)
 
             i = 0
             p = [0, 0, 0, 0, 0, 0]
@@ -1212,7 +1193,7 @@ class DtzTable(Table):
             while i < self.num:
                 piece_type = pc[i] & 0x07
                 color = (pc[i] ^ cmirror) >> 3
-                bb = _bb(board, piece_type, color)
+                bb = board.pieces_mask(piece_type, color)
 
                 square = chess.bit_scan(bb)
                 while square != -1 and square is not None:
@@ -1283,8 +1264,6 @@ class Tablebases(object):
                 if wdl_table.key in self.wdl:
                     self.wdl[wdl_table.key].close()
 
-                wdl_table.init_table_wdl()
-
                 self.wdl[wdl_table.key] = wdl_table
                 self.wdl[wdl_table.mirrored_key] = wdl_table
 
@@ -1295,8 +1274,6 @@ class Tablebases(object):
                 if dtz_table.key in self.dtz:
                     self.dtz[dtz_table.key].close()
 
-                dtz_table.init_table_dtz()
-
                 self.dtz[dtz_table.key] = dtz_table
                 self.dtz[dtz_table.mirrored_key] = dtz_table
 
@@ -1316,7 +1293,7 @@ class Tablebases(object):
         return self.wdl[key].probe_wdl_table(board)
 
     def probe_ab(self, board, alpha, beta):
-        for move in board.generate_pseudo_legal_moves():
+        for move in board.generate_legal_moves():
             # Only look at non-ep captures.
             if not board.piece_type_at(move.to_square):
                 continue
@@ -1324,11 +1301,6 @@ class Tablebases(object):
             # Do the move.
             board.push(move)
 
-            # Only look at legal moves.
-            if board.was_into_check():
-                board.pop()
-                continue
-
             v_plus, success = self.probe_ab(board, -beta, -alpha)
             board.pop()
 
@@ -1374,7 +1346,7 @@ class Tablebases(object):
         -2
         """
         # Positions with castling rights are not in the tablebase.
-        if board.castling_rights != chess.CASTLING_NONE:
+        if board.castling_rights:
             return None
 
         # Probe.
@@ -1390,20 +1362,14 @@ class Tablebases(object):
         v1 = -3
 
         # Look at least at all legal en-passant captures.
-        for move in board.generate_pseudo_legal_moves(castling=False, pawns=True, knights=False, bishops=False, rooks=False, queens=False, king=False):
+        for move in board.generate_legal_moves(castling=False, pawns=True, knights=False, bishops=False, rooks=False, queens=False, king=False):
             # Filter out non-en-passant moves.
-            diff = abs(move.to_square - move.from_square)
-            if not ((diff == 7 or diff == 9) and not board.occupied & chess.BB_SQUARES[move.to_square]):
+            if not board.is_en_passant(move):
                 continue
 
             # Do the move.
             board.push(move)
 
-            # Filter out illegal moves.
-            if board.was_into_check():
-                board.pop()
-                continue
-
             v0_plus, success = self.probe_ab(board, -2, 2)
             board.pop()
 
@@ -1419,21 +1385,10 @@ class Tablebases(object):
             if v1 >= v:
                 v = v1
             elif v == 0:
-                # Check whether there is at least one legal non-ep move.
-                found_move = False
-                for move in board.generate_legal_moves():
-                    if board.piece_type_at(move.from_square) != chess.PAWN:
-                        found_move = True
-                        break
-
-                    diff = abs(move.to_square - move.from_square)
-                    if not ((diff == 7 or diff == 9) and not board.occupied & chess.BB_SQUARES[move.to_square]):
-                        found_move = True
-                        break
-
-                    # If not, then we are forced to play the losing ep capture.
-                    if not found_move:
-                        v = v1
+                # If there is not at least one legal non-en-passant move we are
+                # forced to play the losing en-passant cature.
+                if all(board.is_en_passant(move) for move in board.generate_legal_moves()):
+                    v = v1
 
         return v
 
@@ -1457,16 +1412,12 @@ class Tablebases(object):
             return 1 if wdl == 2 else 101
 
         if wdl > 0:
-            # Generate all legal non capturing pawn moves.
-            for move in board.generate_pseudo_legal_moves(castling=False, pawns=True, knights=False, bishops=False, rooks=False, queens=False, king=False):
-                diff = abs(move.to_square - move.from_square)
-                if diff == 7 or diff == 9:
+            # Generate all legal non-capturing pawn moves.
+            for move in board.generate_legal_moves(castling=False, pawns=True, knights=False, bishops=False, rooks=False, queens=False, king=False):
+                if board.is_capture(move):
                     continue
 
                 board.push(move)
-                if board.was_into_check():
-                    board.pop()
-                    continue
 
                 v_plus, success = self.probe_ab(board, -2, -wdl + 1)
                 board.pop()
@@ -1490,14 +1441,11 @@ class Tablebases(object):
         if wdl > 0:
             best = 0xffff
 
-            for move in board.generate_pseudo_legal_moves(pawns=False):
+            for move in board.generate_legal_moves(pawns=False):
                 if board.piece_type_at(move.to_square) != chess.NONE:
                     continue
 
                 board.push(move)
-                if board.was_into_check():
-                    board.pop()
-                    continue
 
                 v_plus = self.probe_dtz(board)
                 board.pop()
@@ -1514,11 +1462,8 @@ class Tablebases(object):
         else:
             best = -1
 
-            for move in board.generate_pseudo_legal_moves():
+            for move in board.generate_legal_moves():
                 board.push(move)
-                if board.was_into_check():
-                    board.pop()
-                    continue
 
                 if board.halfmove_clock == 0:
                     if wdl == -2:
@@ -1580,16 +1525,12 @@ class Tablebases(object):
 
         v1 = -3
 
-        for move in board.generate_pseudo_legal_moves(castling=False, pawns=True, knights=False, bishops=False, rooks=False, queens=False, king=False):
+        for move in board.generate_legal_moves(castling=False, pawns=True, knights=False, bishops=False, rooks=False, queens=False, king=False):
             # Filter out non-en-passant moves.
-            diff = abs(move.to_square - move.from_square)
-            if not ((diff == 7 or diff == 9) and not board.occupied & chess.BB_SQUARES[move.to_square]):
+            if not board.is_en_passant(move):
                 continue
 
             board.push(move)
-            if board.was_into_check():
-                board.pop()
-                continue
 
             v0_plus, success = self.probe_ab(board, -2, 2)
             board.pop()
@@ -1619,18 +1560,7 @@ class Tablebases(object):
             elif v1 >= 0:
                 v = v1
             else:
-                found_move = False
-                for move in board.generate_legal_moves():
-                    if board.piece_type_at(move.from_square) != chess.PAWN:
-                        found_move = True
-                        break
-
-                    diff = abs(move.to_square - move.from_square)
-                    if not ((diff == 7 or diff == 9) and not board.occupied & chess.BB_SQUARES[move.to_square]):
-                        found_move = True
-                        break
-
-                if not found_move:
+                if all(board.is_en_passant(move) for move in board.generate_legal_moves()):
                     v = v1
 
         return v
@@ -1648,5 +1578,5 @@ class Tablebases(object):
     def __enter__(self):
         return self
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, exc_type, exc_value, traceback):
         self.close()
