@@ -22,6 +22,7 @@ import signal
 import subprocess
 import logging
 import threading
+import copy
 
 try:
     import queue
@@ -75,11 +76,11 @@ class OptionMap(collections.MutableMapping):
 
     def __eq__(self, other):
         for key, value in self.items():
-            if not key in other or other[key] != value:
+            if key not in other or other[key] != value:
                 return False
 
         for key, value in other.items():
-            if not key in self or self[key] != value:
+            if key not in self or self[key] != value:
                 return False
 
         return True
@@ -99,6 +100,7 @@ class InfoHandler(object):
         self.info["refutation"] = {}
         self.info["currline"] = {}
         self.info["pv"] = {}
+        self.info["score"] = {}
 
     def depth(self, x):
         """Received search depth in plies."""
@@ -126,7 +128,12 @@ class InfoHandler(object):
         self.info["pv"][self.info.get("multipv", 1)] = moves
 
     def multipv(self, num):
-        """Received a new multipv number, starting at 1."""
+        """
+        Received a new multipv number, starting at 1.
+
+        If multipv occurs in an info line, this is guaranteed to be called
+        before *score* or *pv*.
+        """
         self.info["multipv"] = num
 
     def score(self, cp, mate, lowerbound, upperbound):
@@ -140,11 +147,19 @@ class InfoHandler(object):
 
         lowerbound and upperbound are usually *False*. If *True*, the sent
         score are just a lowerbound or upperbound.
+
+        In MultiPV mode this is related to the most recent *multipv* number
+        sent by the engine.
         """
-        self.info["score"] = Score(cp, mate, lowerbound, upperbound)
+        self.info["score"][self.info.get("multipv", 1)] = Score(cp, mate, lowerbound, upperbound)
 
     def currmove(self, move):
-        """Received a move the engine is currently thinking about."""
+        """
+        Received a move the engine is currently thinking about.
+
+        These moves come directly from the engine. So the castling move
+        representation depends on the UCI_Chess960 option of the engine.
+        """
         self.info["currmove"] = move
 
     def currmovenumber(self, x):
@@ -204,6 +219,7 @@ class InfoHandler(object):
         order to keep the locking in tact.
         """
         self.lock.acquire()
+        self.info.pop("multipv", None)
 
     def post_info(self):
         """
@@ -214,19 +230,15 @@ class InfoHandler(object):
         """
         self.lock.release()
 
-    def pre_bestmove(self, line):
-        """A new bestmove command is about to be processed."""
-        pass
-
     def on_bestmove(self, bestmove, ponder):
         """A new bestmove and pondermove have been received."""
         pass
 
-    def post_bestmove(self):
+    def on_go(self):
         """
-        A new bestmove command was processed.
+        A go command is being sent.
 
-        Since this indicates that the current search has been finished the
+        Since information about the previous search is invalidated the
         dictionary with the current information will be cleared.
         """
         with self.lock:
@@ -234,6 +246,7 @@ class InfoHandler(object):
             self.info["refutation"] = {}
             self.info["currline"] = {}
             self.info["pv"] = {}
+            self.info["score"] = {}
 
     def acquire(self, blocking=True):
         return self.lock.acquire(blocking)
@@ -245,7 +258,7 @@ class InfoHandler(object):
         self.acquire()
         return self.info
 
-    def __exit__(self, type, value, traceback):
+    def __exit__(self, exc_type, exc_value, traceback):
         self.release()
 
 
@@ -366,9 +379,13 @@ class SetOptionCommand(IsReadyCommand):
     def __init__(self, options):
         super(SetOptionCommand, self).__init__()
 
+        self.uci_chess960 = None
         self.option_lines = []
 
         for name, value in options.items():
+            if name.lower() == "uci_chess960":
+                self.uci_chess960 = value
+
             builder = []
             builder.append("setoption name ")
             builder.append(name)
@@ -388,6 +405,9 @@ class SetOptionCommand(IsReadyCommand):
         for option_line in self.option_lines:
             engine.send_line(option_line)
 
+        if self.uci_chess960 is not None:
+            engine.uci_chess960 = self.uci_chess960
+
         super(SetOptionCommand, self).execute(engine)
 
 
@@ -401,32 +421,53 @@ class PositionCommand(IsReadyCommand):
     def __init__(self, board):
         super(PositionCommand, self).__init__()
 
+        self.board = copy.deepcopy(board)
+
+    def execute(self, engine):
         builder = []
         builder.append("position")
 
+        # Take back moves to obtain the first FEN we know. Later giving the
+        # moves explicitly allows for transposition detection.
         switchyard = collections.deque()
-        while board.move_stack:
-            switchyard.append(board.pop())
+        while self.board.move_stack:
+            switchyard.append(self.board.pop())
 
-        fen = board.fen()
-        if fen == chess.STARTING_FEN:
+        # Validate castling rights.
+        if not engine.uci_chess960:
+            standard_chess_status = self.board.status(allow_chess960=False)
+            chess960_status = self.board.status(allow_chess960=True)
+            if standard_chess_status & chess.STATUS_BAD_CASTLING_RIGHTS and not chess960_status & chess.STATUS_BAD_CASTLING_RIGHTS:
+                LOGGER.error("%s not in UCI_Chess960 mode but position has non-standard castling rights", engine.process)
+
+                # Just send the final FEN without transpositions in hopes that
+                # this will work.
+                while switchyard:
+                    self.board.push(switchyard.pop())
+
+        # Send startposition.
+        if self.board.fen() == chess.STARTING_FEN:
             builder.append("startpos")
         else:
             builder.append("fen")
-            builder.append(fen)
 
+            if engine.uci_chess960:
+                builder.append(self.board.shredder_fen())
+            else:
+                builder.append(self.board.fen())
+
+        # Send moves.
         if switchyard:
             builder.append("moves")
 
             while switchyard:
                 move = switchyard.pop()
-                builder.append(move.uci())
-                board.push(move)
+                builder.append(self.board.uci(move, chess960=engine.uci_chess960))
+                self.board.push(move)
 
-        self.buf = " ".join(builder)
+        engine.board = self.board
+        engine.send_line(" ".join(builder))
 
-    def execute(self, engine):
-        engine.send_line(self.buf)
         super(PositionCommand, self).execute(engine)
 
 
@@ -481,18 +522,26 @@ class GoCommand(Command):
         if infinite:
             builder.append("infinite")
 
-        if searchmoves:
-            builder.append("searchmoves")
-            for move in searchmoves:
-                builder.append(move.uci())
+        self.searchmoves = searchmoves
 
         self.buf = " ".join(builder)
 
     def execute(self, engine):
+        for info_handler in engine.info_handlers:
+            info_handler.on_go()
+
+        # Append searchmoves last. They can not be built beforehand because
+        # they also depend on the UCI_Chess960 option.
+        builder = [self.buf]
+        if self.searchmoves:
+            builder.append("searchmoves")
+            for move in self.searchmoves:
+                builder.append(engine.board.uci(move, chess960=engine.uci_chess960))
+
         engine.bestmove = None
         engine.ponder = None
         engine.bestmove_received.clear()
-        engine.send_line(self.buf)
+        engine.send_line(" ".join(builder))
         if self.infinite or self.ponder:
             self.set_result(None)
         else:
@@ -736,6 +785,9 @@ class Engine(object):
     def __init__(self, process):
         self.process = process
 
+        self.board = chess.Board()
+        self.uci_chess960 = None
+
         self.name = None
         self.author = None
         self.options = OptionMap()
@@ -826,36 +878,47 @@ class Engine(object):
             return
 
     def _uciok(self):
+        # Set UCI_Chess960 default value.
+        if self.uci_chess960 is None and "UCI_Chess960" in self.options:
+            self.uci_chess960 = self.options["UCI_Chess960"].default
+
         self.uciok.set()
 
     def _readyok(self):
         self.readyok.set()
 
     def _bestmove(self, arg):
-        for info_handler in self.info_handlers:
-            info_handler.pre_bestmove(arg)
-
         tokens = arg.split(None, 2)
 
+        self.bestmove = None
         if tokens[0] != "(none)":
-            self.bestmove = chess.Move.from_uci(tokens[0])
-        else:
-            self.bestmove = None
+            try:
+                self.bestmove = self.board.parse_uci(tokens[0])
+            except ValueError:
+                LOGGER.exception("exception parsing bestmove")
 
-        if len(tokens) >= 3 and tokens[1] == "ponder" and tokens[2] != "(none)":
-            self.ponder = chess.Move.from_uci(tokens[2])
-        else:
-            self.ponder = None
+        self.ponder = None
+        if self.bestmove is not None and len(tokens) >= 3 and tokens[1] == "ponder" and tokens[2] != "(none)":
+            # The ponder move must be legal after the bestmove. Generally we
+            # trust the engine on this. But we still have to convert
+            # non-UCI_Chess960 castling moves.
+            try:
+                self.ponder = chess.Move.from_uci(tokens[2])
+                if self.ponder.from_square in (chess.E1, chess.E8) and self.ponder.to_square in (chess.C1, chess.C8, chess.G1, chess.G8):
+                    # Make a copy of the board to avoid race conditions.
+                    board = copy.deepcopy(self.board)
+                    board.push(self.bestmove)
+                    board.push(self.ponder)
+                    self.ponder = self.board.parse_uci(tokens[2])
+            except ValueError:
+                LOGGER.exception("exception parsing bestmove ponder")
+                self.ponder = None
 
         self.bestmove_received.set()
 
         for info_handler in self.info_handlers:
             info_handler.on_bestmove(self.bestmove, self.ponder)
 
-        for info_handler in self.info_handlers:
-            info_handler.post_bestmove()
-
-
     def _copyprotection(self, arg):
         # TODO: Implement copyprotection
         pass
@@ -868,11 +931,12 @@ class Engine(object):
         if not self.info_handlers:
             return
 
+        # Notify info handlers of start.
         for info_handler in self.info_handlers:
             info_handler.pre_info(arg)
 
-        current_parameter = None
-
+        # Initialize parser state.
+        board = None
         pv = None
         score_kind = None
         score_cp = None
@@ -913,6 +977,7 @@ class Engine(object):
             try:
                 intval = int(token)
             except ValueError:
+                LOGGER.exception("exception parsing integer token")
                 return
 
             for info_handler in self.info_handlers:
@@ -922,11 +987,26 @@ class Engine(object):
             try:
                 move = chess.Move.from_uci(token)
             except ValueError:
+                LOGGER.exception("exception parsing move token")
                 return
 
             for info_handler in self.info_handlers:
                 fn(info_handler, move)
 
+        # Find multipv parameter first.
+        if "multipv" in arg:
+            current_parameter = None
+            for token in arg.split(" "):
+                if token == "string":
+                    break
+
+                if current_parameter == "multipv":
+                    handle_integer_token(token, lambda handler, val: handler.multipv(val))
+
+                current_parameter = token
+
+        # Parse all other parameters.
+        current_parameter = None
         for token in arg.split(" "):
             if current_parameter == "string":
                 string.append(token)
@@ -947,6 +1027,9 @@ class Engine(object):
 
                 if current_parameter == "pv":
                     pv = []
+
+                if current_parameter in ("refutation", "pv", "currline"):
+                    board = copy.deepcopy(self.board)
             elif current_parameter == "depth":
                 handle_integer_token(token, lambda handler, val: handler.depth(val))
             elif current_parameter == "seldepth":
@@ -957,11 +1040,12 @@ class Engine(object):
                 handle_integer_token(token, lambda handler, val: handler.nodes(val))
             elif current_parameter == "pv":
                 try:
-                    pv.append(chess.Move.from_uci(token))
+                    pv.append(board.push_uci(token))
                 except ValueError:
-                    pass
+                    LOGGER.exception("exception parsing pv")
             elif current_parameter == "multipv":
-                handle_integer_token(token, lambda handler, val: handler.multipv(val))
+                # Ignore multipv. It was already parsed before anything else.
+                pass
             elif current_parameter == "score":
                 if token in ("cp", "mate"):
                     score_kind = token
@@ -973,12 +1057,12 @@ class Engine(object):
                     try:
                         score_cp = int(token)
                     except ValueError:
-                        pass
+                        LOGGER.exception("exception parsing score cp value")
                 elif score_kind == "mate":
                     try:
                         score_mate = int(token)
                     except ValueError:
-                        pass
+                        LOGGER.exception("exception parsing score mate value")
             elif current_parameter == "currmove":
                 handle_move_token(token, lambda handler, val: handler.currmove(val))
             elif current_parameter == "currmovenumber":
@@ -994,19 +1078,19 @@ class Engine(object):
             elif current_parameter == "refutation":
                 try:
                     if refutation_move is None:
-                        refutation_move = chess.Move.from_uci(token)
+                        refutation_move = board.push_uci(token)
                     else:
-                        refuted_by.append(chess.Move.from_uci(token))
+                        refuted_by.append(board.push_uci(token))
                 except ValueError:
-                    pass
+                    LOGGER.exception("exception parsing refutation")
             elif current_parameter == "currline":
                 try:
                     if currline_cpunr is None:
                         currline_cpunr = int(token)
                     else:
-                        currline_moves.append(chess.Move.from_uci(token))
+                        currline_moves.append(board.push_uci(token))
                 except ValueError:
-                    pass
+                    LOGGER.exception("exception parsing currline")
 
         end_of_parameter()
 
@@ -1014,6 +1098,7 @@ class Engine(object):
             for info_handler in self.info_handlers:
                 info_handler.string(" ".join(string))
 
+        # Notify info handlers of end.
         for info_handler in self.info_handlers:
             info_handler.post_info()
 
@@ -1052,12 +1137,12 @@ class Engine(object):
                 try:
                     min = int(token)
                 except ValueError:
-                    pass
+                    LOGGER.exception("exception parsing option min")
             elif current_parameter == "max":
                 try:
                     max = int(token)
                 except ValueError:
-                    pass
+                    LOGGER.exception("exception parsing option max")
             elif current_parameter == "default":
                 default.append(token)
             elif current_parameter == "var":
@@ -1080,6 +1165,7 @@ class Engine(object):
             try:
                 default = int(default)
             except ValueError:
+                LOGGER.exception("exception parsing option spin default")
                 default = None
 
         option = Option(" ".join(name), type, default, min, max, var)
